//<copyright>
// 
// Copyright (c) 1993,94
// Institute for Information Processing and Computer Supported New Media (IICM),
// Graz University of Technology, Austria.
// 
//</copyright>

//<file>
//
// Name:        baseprot.h
//
// Purpose:     class BaseProt (basic protocol handling class) 
//              and derived classes
//
// Created:     10 Oct 94    Thomas Dietinger
//
// Modified:    29 Mar 95    Thomas Dietinger
//
//
//
// Description:
//
//
//</file>

#ifndef HG_HYPERG_BASEPROT_H
#define HG_HYPERG_BASEPROT_H

#include <InterViews/boolean.h>
#include <InterViews/resource.h>
#include <utils/str.h>
#include <hyperg/object.h>

//
// bitmask that tells us the features a protocol library has implemented
//
extern const unsigned long identFunc;
extern const unsigned long readFunc;
extern const unsigned long writeFunc;
extern const unsigned long moveFunc;

extern const unsigned long copyFunc;
extern const unsigned long eraseFunc;
extern const unsigned long unlinkFunc;
extern const unsigned long keyQueryFunc;

extern const unsigned long fTQueryFunc;
extern const unsigned long childrenFunc;
extern const unsigned long parentsFunc;
extern const unsigned long linksToFunc;

extern const unsigned long linksFromFunc;
extern const unsigned long lockFunc;
extern const unsigned long unlockFunc;
extern const unsigned long chgPasswdFunc;

//
// types that specfiy what we want when we do an 'openObj'
//
enum OpenMode { getDataMode, putDataMode, getChildrenMode, getParentsMode, 
                getLinksMode, getReferencesMode, keyQueryMode, fullTextQueryMode };

class ObjectList;

void splitURL(RString url, RString& protocol, RString& user, RString& passwd,
              RString& host, int portNr, RString& path);

/////////////////////////////////////////////////////////////////
// class BaseFD: basic file descriptor and info class (dummy implementation)
class BaseFD
{
public:
    BaseFD();
    virtual ~BaseFD() {};

    OpenMode mode_;
    RString Type_;          // file descriptor type, to be sure to cast correctly
    long bufLen_;           // number of bytes for data transmission or
                            // number of elements for lists to fetch or put at once
    long pos_;
    long dataLen_;          // complete length of data (or -1 if unknown)
                            // either # of bytes or
                            //        # of list elements
    ObjectPtr objPtr_;      // may be used or not

    // file descriptor handle representation e.g:
    // SOCKET fd_; or HANDLE fd_; or FILE* fd_; ...
};

/////////////////////////////////////////////////////////////////
// class BaseProt: basic protocol class (dummy implementation)

class BaseProt : public Resource
{
public:
    BaseProt(RString connectStr, 
             boolean connect = false);  // usual way
    BaseProt(BaseProt* parentProt);     // if we want to use another (already used) protocol
    virtual ~BaseProt();

    // get object information by destination ID/URL
    virtual Object getObjectInfo(RString dest)          { return ""; }

    virtual boolean reconnect(RString connectStr, 
                              boolean connect = false)  { return false; }

    virtual RString getProtType() const                 { return protType_; }
    virtual RString getHost() const                     { return host_; }
    virtual int     getPortNr() const                   { return portNr_; }
    virtual RString getHomePath() const                 { return homePath_; }
    virtual boolean getConnectState() const             { return connect_; }

    virtual RString getUser()                           { return user_; }
    virtual RString getPassWord()                       { return passWord_; }
    virtual RString getConnectStr()                     { return connectStr_; }

    virtual boolean setUser(RString user)               { user_ = user; return true; }
    virtual boolean setPassWord(RString pw)             { passWord_ = pw; return true; }

    virtual void setErrorCode(unsigned int nId);

    virtual unsigned long features()                    { return features_; }

    //
    // placeholders for protocol functions
    //

    // check to see whether we can use this connection type with this protocol
    virtual boolean     useConnection(RString connectStr)       { return false; }

    virtual boolean     identify(const RString& name, 
                                 const RString& passwd = "", 
                                 boolean encrypt = false)       { return false; }
    virtual boolean     changePassword(const RString& oldPassword, 
                                       const RString& newPassword)    { return false; }

    virtual BaseFD*     open(ObjectPtr objPtr,           // pointer to object of interest
                             OpenMode mode,              // opening mode
                             ObjectList* objList = 0,    // pointer to additional list (e.g. list of activated collections)
                             RString helpStr = ""        // additional string (e.g. query string)
                             )                              { return 0; }
    virtual boolean     close(BaseFD* fd)                   { return false; }

    virtual long        getData(BaseFD* fd, 
                                char* dataBuf, 
                                long buflen)                { return -1; }
    virtual long        putData(BaseFD* fd, 
                                char* dataBuf, 
                                long buflen)                { return -1; }

    virtual ObjectList* getList(BaseFD* fd, 
                                long count,
                                ObjectList* oldList = 0)    { return nil; }
    virtual boolean     putList(BaseFD* fd, 
                                ObjectList* newObjs)        { return false; }

    virtual boolean     insert(ObjectPtr object, 
                               ObjectPtr toColl)            { return false; }

    virtual boolean     erase(ObjectPtr)                    { return false; }
    virtual boolean     moveTo(ObjectPtr, 
                               ObjectPtr fromColl, 
                               ObjectPtr toColl)            { return false; }
    virtual boolean     copyTo(ObjectPtr,
                               ObjectPtr toColl)            { return false; }
    virtual boolean     unlinkFrom(ObjectPtr, 
                                   ObjectPtr fromColl)      { return false; }

    virtual HGObject*   lock(ObjectPtr)                     { return nil; }
    virtual boolean     unlock(ObjectPtr)                   { return false; }

    virtual boolean     modifyObjAttribute(ObjectPtr objPtr,    // referred object
                                           RString oldAttrLine, // old attribute
                                           RString newAttrLine  // new attribute
                                           )                { return false; }

protected:

    // protocol attributes (may be used or not)
    boolean connect_;       // connection state
    RString protType_;      // protocol type
    RString host_;          // the host we are connected to
    int     portNr_;        // the used TCP port number
    RString user_;          // user-id if any
    RString passWord_;      // password of user if any
    boolean encrypt_;       // encrypted password ?
    RString homePath_;      // home path if any

    RString connectStr_;    // the complete connection string

    int errCode_;           // error code, if any or 0
    RString errStr_;        // error string, if any or ""

    unsigned long features_;    // protocol features, set by constructor
}; // BaseProt


#endif
