// This may look like C code, but it is really -*- C++ -*-

//<copyright>
// 
// Copyright (c) 1993
// Institute for Information Processing and Computer Supported New Media (IICM),
// Graz University of Technology, Austria.
// 
//</copyright>

//<file>
//
// Name:       userprof.h
//
// Purpose:    UserProfile - manages the user record
//
// Created:    07 Nov 93    Frank Kappe
//
// Modified:   07 Nov 93    Frank Kappe
//
//
//
// Description:
//
// Contains class UserProfile. This class is not used at all by other parts of 
// the 'apps' library. Rather, is is used by applications (clients and gateways).
//
//</file>

#ifndef hg_apps_userprof_h
#define hg_apps_userprof_h

#include "hgobject.h"

//<class>
//
// Name:       UserProfile
//
// Purpose:    Manages the user profile on the client side.
//
// Public Interface:
//
// - UserProfile(HGStub* db)
//   Creates a user profile from the current connection. The connection must have
//   been opened before, and remains the same during the whole session.
//
// - void initProfile()
//   (Re-)initializes the local copy of the user profile from information stored
//   in the database.
//
// - ObjectPtr homeCollection(const RString& defaultname)
//   Returns the user's home collection, if one is set. Otherwise, return the 
//   collection specified by 'defaultname'. If all fails, return nil.
//
// - boolean changePassword(const RString& old, const RString& new)
//   Allows to change one of the passwords associated with this user. The old
//   password has to be supplied for this operation, as well as the new one.
//   Return false on failure, in which case the error() method of the database
//   connection can be examied to determine the cause of the error. If error()
//   reports no error but changePassword() returns false, the old password was
//   wrong. 
//
// - boolean changePasswordCrypt(const RString& old, const RString& new)
//   Same as changePassword() but allows to specify an already encrypted version
//   of the password (e.g. taken from the system password file). The old password
//   has to be supplied in clear text (!) for security reasons.
//
// - static UserProfile& instance();
//   Returns an instance of UserProfile. Creates it (only) when called for the 
//   first time.
//
// Private Methods:
//
// Note: The "enCrypt" and "validatePassword" methods are NOT used to really
//       validate user access to the Hyper-G server. They are only used when
//       changing passwords. Of course, encryption must use the same algorithm
//       as the server.
//
//</class>

class UserProfile 
{
public:
    UserProfile(HgStub* db);
    ~UserProfile();
    void initProfile();
    static void resetProfile();

    ObjectPtr homeCollection(const RString& defaultname);
    boolean changePassword(const RString& oldpw, const RString& newpw);
    boolean changePasswordCrypt(const RString& oldpw, const RString& newpw);

    static UserProfile& instance();

    static UserProfile* instance_;

private:
    RString enCrypt(const RString& clear);
    boolean validatePassword(const RString& clear, const RString& encr);

    HgStub* db_;
    Object  o_;  

    ObjectPtr homecollection_;
    RString oldHome_;
};

extern HgStub* HG;

inline UserProfile& UserProfile::instance() 
{
    if (instance_ == nil) 
    {
        instance_ = new UserProfile(HG);
    }
    return *instance_;
}

inline void UserProfile::resetProfile() 
{
    if (instance_ != nil) 
    {
        delete instance_;
        instance_ = new UserProfile(HG);
    }
    return;
}

#endif
