//<copyright>
// 
// Copyright (c) 1993,94,95
// Institute for Information Processing and Computer Supported New Media (IICM),
// Graz University of Technology, Austria.
// 
//</copyright>

//<file>
//
// Name:        vecutil.C
//
// Purpose:     utilities for 3D vectors
//
// Created:     15 Mar 93   Michael Pichler
//
// Changed:      8 May 95   Michael Pichler
//
//
//</file>


#include "vecutil.h"

#include <ge3d/ge3d.h>
#include <math.h>


// became obsolete by ge3dPushIdentity
// const float identity_matrix [4][4] =
// { { 1.0, 0.0, 0.0, 0.0 },
//   { 0.0, 1.0, 0.0, 0.0 },
//   { 0.0, 0.0, 1.0, 0.0 },
//   { 0.0, 0.0, 0.0, 1.0 }
// };


// rayhitscube
// tests wheter ray A + b.t hits the (axis aligned) cube with vertices (min, max)
// (min.x < max.x etc. required) at a value of t, t > tnear and t < tmin, from outside;
// if so, t is returned (but tmin is not changed), 0.0 otherwise
// (this test requires only 3 divisions, 9 multiplications and some additions/comparisons)

float rayhitscube (const point3D& A, const vector3D& b, float tnear, float tmin,
                   const point3D& min, const point3D& max)
{
  float thit;
  point3D hp;  // hit point

  if (b.x)
  { thit = ((b.x > 0 ? min.x : max.x) - A.x) / b.x;
    if (thit > tnear && thit < tmin)
    { // check if hit point hp = A + thit * b  lies within face
      pol3D (A, thit, b, hp);
      if (min.y <= hp.y && hp.y <= max.y && min.z <= hp.z && hp.z <= max.z)
        return thit;
    }
  }

  if (b.y)
  { thit = ((b.y > 0 ? min.y : max.y) - A.y) / b.y;
    if (thit > tnear && thit < tmin)
    { // check if hit point hp = A + thit * b  lies within face
      pol3D (A, thit, b, hp);
      if (min.x <= hp.x && hp.x <= max.x && min.z <= hp.z && hp.z <= max.z)
        return thit;
    }
  }

  if (b.z)
  { thit = ((b.z > 0 ? min.z : max.z) - A.z) / b.z;
    if (thit > tnear && thit < tmin)
    { // check if hit point hp = A + thit * b  lies within face
      pol3D (A, thit, b, hp);
      if (min.x <= hp.x && hp.x <= max.x && min.y <= hp.y && hp.y <= max.y)
        return thit;
    }
  }

  return 0.0;

} // rayhitscube



// invertmatrix
// overwrites 4 by 4 matrix a with its inverse
// (algorithm from: Schwarz, "Numerische Mathematik")
// (row permutations for less numerical errors)

void invertmatrix (matrix4D a)
{
  int n = 4;

  float max, s, h, q, pivot;
  int p [4];
  int i, j, k;

  for (k = 0; k < n; k++)
  { max = 0.0;
    p [k] = 0;

    for (i = k; i < n; i++)
    { s = 0.0;
      for (j = k; j < n; j++)
        s += fabs (a [i][j]);
      q = fabs (a [i][k]) / s;
      if (q > max)
      { max = q;  
        p [k] = i;
      }
    } // for i

    if (!max)
      return;
    if (p [k] != k)
      for (j = 0; j < n; j++)
      { h = a [k][j];  a [k][j] = a [p[k]][j];  a [p[k]][j] = h;
      }

    pivot = a [k][k];
    for (j = 0; j < n; j++)
      if (j != k)
      { a [k][j] = - a [k][j] / pivot;
        for (i = 0; i < n; i++)
          if (i != k)
            a [i][j] += a [i][k] * a [k][j];
      } // for j, if

    for (i = 0; i < n; i++)
      a [i][k] /= pivot;
    a [k][k] = 1 / pivot;

  } // for k

  for (k = n-2; k >= 0; k--)
    if (p [k] != k)
      for (i = 0; i < n; i++)
      { h = a [i][k];  a [i][k] = a [i][p[k]];  a [i][p[k]] = h;
      } // for k, if, for i

} // invertmatrix



// conversion cartesian - spherepoint

// cartes2sphere
// convert cartesian to sphere coordinates

void cartes2sphere (const point3D& p, spherepoint& s)
{
  s.r = sqrt (dot3D (p, p));  // length
  s.phi = atan2 (p.y, p.x);  // angle from x to y axis
  s.theta = (s.r > 0.0) ? acos (p.z / s.r) : 0.0;  // angle from z axis to x-y-plane
}


// sphere2cartes
// convert sphere to cartesian coordinates
// p = (0, 0, r) . ROTy (theta) . ROTz (phi)

void sphere2cartes (const spherepoint& s, point3D& p)
{
  double sinth = sin (s.theta);
  p.x = s.r * cos (s.phi) * sinth;
  p.y = s.r * sin (s.phi) * sinth;
  p.z = s.r * cos (s.theta);
}



// convert radians to degrees
#define DEGREES(R)  ( (R) * (180 / M_PI) )


// rotatearoundaxis
// preconcatenate the current transformation matrix (of ge3d) with one that does a
// rotation about an arbitrary axis u through origin about an angle omega -
// in radians; positive direction = CCW when looking along u towards origin

// void rotatearoundaxis (const vector3D& u, float omega)
// {
//   spherepoint s;
//   cartes2sphere (u, s);

// // TEST !!!
// static int firstcall = 1;

//   // sorry, ge3d (as gl) expects angles in degrees
//   // note: the matrices are PREconcatenated, therefore to be read from bottom up!
// if (firstcall)  cerr << "rotz" << endl;
//   ge3d_rotate_axis ('z', DEGREES (s.phi));
// if (firstcall)  ge3d_print_cur_matrix (), cerr << "roty" << endl;
//   ge3d_rotate_axis ('y', DEGREES (s.theta));
// if (firstcall)  ge3d_print_cur_matrix (), cerr << "rotz" << endl;
//   ge3d_rotate_axis ('z', DEGREES (omega));  // u lies in positive z axis
// if (firstcall)  ge3d_print_cur_matrix (), cerr << "roty" << endl;
//   ge3d_rotate_axis ('y', DEGREES (-s.theta));
// if (firstcall)  ge3d_print_cur_matrix (), cerr << "rotz" << endl;
//   ge3d_rotate_axis ('z', DEGREES (-s.phi));
// if (firstcall)  ge3d_print_cur_matrix ();
// firstcall = 0;
// }


// object to world bounding box

void computeBoundingbox (const point3D& omin, const point3D& omax, point3D& wmin, point3D& wmax)
{
  // transform the bounding box of the object into world coordinates
  // and then put another (aligned) bounding box around - it is exact for
  // cubes but somewhat larger for "round" objects
  // computing the 'best' world bounding box is not worthwile

  // transform all eight vertices of the object bounding and compare it to min/max

  emptyBoundingbox (wmin, wmax);

  point3D p;

  for (int i = 0;  i < 8;  i++)
  {
    ge3d_transform_mc_wc (
      (i & 4) ? omin.x : omax.x,  (i & 2) ? omin.y : omax.y,  (i & 1) ? omin.z : omax.z,
      &p.x, &p.y, &p.z
    );
    extendBoundingbox (p, p, wmin, wmax);
  }

} // computeBoundingbox


// ostream << point3D
// also handles ostream << vector3D

ostream& operator << (ostream& os, const point3D& pt)
{
  os << '(' << pt.x << ", " << pt.y << ", " << pt.z << ')';
  return os;
}


// ostream << point2D

ostream& operator << (ostream& os, const point2D& pt)
{
  os << '(' << pt.x << ", " << pt.y << ')';
  return os;
}


// ostream << spherepoint

ostream& operator << (ostream& os, const spherepoint& sp)
{
  os << '[' << sp.r << ", " << sp.phi << ", " << sp.theta << ']';
  return os;
}


// ostream << matrix4D

ostream& operator << (ostream& os, const matrix4D mat)
{
  os << '\n';

  for (int i = 0;  i < 4;  i++)
  {
    const float* row = &(mat [i][0]);
    os << row [0] << ", " << row [1] << ", " << row [2] << ", " << row [3] << endl;
  }

  return os << endl;
}
